#ifndef instancemap_include_declared
#define instancemap_include_declared

#include "Instance.h"
#include <algorithm> // 

typedef list<Instance*> InstanceList;
typedef list<Instance*>::iterator InstanceIterator; 

class InstanceMap {
 private:
  int _numberOfInstances;
  int _iterCounter;
  InstanceList *_pInstanceList;
  InstanceIterator ii; //
  // for motif detection (of fixed width !)
  int *_pW; //
  StringVector*** _pInstanceCounts; //
  StringVector::iterator specIter;
  
 public:
  // constructor
  InstanceMap(); //
  InstanceMap(int *pW); // new
 
  // destructor
  ~InstanceMap(); //

  // local counts //
  void AddInstanceToCounts(Instance * pInst); // new
  void RemoveInstanceFromCounts(Instance * pInst); // new
  StringVector*** GetInstanceCounts(){return _pInstanceCounts;}; // new
  int** GetIntegerCounts(); // new
  
  // inspectors
  int NumberOfInstances(){return _numberOfInstances;};
  list<Instance *>::iterator begin(){return _pInstanceList->begin();};
  list<Instance *>::iterator end(){return _pInstanceList->end();};
  int * GetMotifWidth(){ return _pW;}; // new nodig ?

  // adaptors
  InstanceMap * AddInstance(Instance *pInst); //
  void AddInstances(InstanceList *pList); // new nodig ?
  InstanceMap * RemoveInstance(Instance *pInst); //
  InstanceMap * ReplaceInstance(Instance *oldInst, Instance *newInst); //
  InstanceMap * RemoveInstanceFromSequence(SequenceObject *pSeq); //
  InstanceMap * Erase(list<Instance *>::iterator iter); //
  InstanceMap * ExtendLeft(int pos); 
  InstanceMap * ExtendRight(int pos); 
  void ClearMap(); //

};

#endif
