#ifndef instanceset_include_declared
#define instanceset_include_declared

#include "Instance.h"
#include "PWM.h"

class CompareID{
  public:
  bool operator()(Instance *Inst1, Instance *Inst2) //Inst1 is the instance to be inserted.
  {  
    int c1 = (*(Inst1->PrintSeqName())).compare(*(Inst2->PrintSeqName()));
    // sort by seq-name (alfabetically)
    if (c1 < 0) return true; // insert the instance1 with lower alfabetic sequence
    if (c1 == 0) // same sequence
    { 
      // sort by strand (first + then -)
      if ( Inst1->Strand() != Inst2->Strand()) // different strand 
      {
        if (Inst1->Strand() == plus_strand) return true; // insert + strand 
        else return false; // do not insert - strand here
      }
      else // same strand
      {
        // during LoadInstances()-phase - InstanceSet::StoreInstance() : _iStore == 0
        if ( Inst1->IntStore() == 0)
        {  
          // 2011/09/18
          // store all instances of same sequence/strand in one artificial
          // this is to avoid errors/biasses due to the order in which instances are loaded 
          return false; 
        }
        // during final add-phase InstanceSet::SplitOverlappingInstances() - _iStore == 1;
        else //if ( Inst1->IntStore() == 1)
        {  
          // store in ascending order, do not look at overlap
          if ( Inst2->Start() > Inst1->Start() ) return true; // insert preceeding
          //if ( Inst1->Start() > Inst2->Start() ) return false; // do not insert following here
          // if ( Inst2->Start() == Inst1->Start()) // should not be the case after the splitting up
          return false;  
        }
      }
    }
    return false; //do not insert higher alfabetic sequence here
  }
};

typedef set<Instance*, CompareID> iSet;
typedef set<Instance*>::iterator iSetIter;

class InstanceSet{

  private:
  // variables
  iSet * _pSet;
  iSetIter iter;
  int _nbrNodes;
  int _nbrSeq; // S
  double** _pSeqVolume;
  string * _pError;

  public:
  // constructor
  InstanceSet();

  // destructor
  ~InstanceSet();
 
  // input 
  void StoreInstance(Instance* inst);

  // adaptor
  void SetpSet(iSet * set){_pSet = set; return;}; // for unlinking purposes
  void NormalizeWeights(const double &norm);
  void GroupExactInstances();//FC
  void SplitOverlappingInstances();//FC
  string RemoveLowProbIndividuals(const int &datasize, const double &cut);
  void AssignNodesToSeqs(); 
  // other functions
  double * BuildAsymmetricalMatrix(vector<string*> * pAllSets);

  // inspector
  int NbrSeq(){return _nbrSeq;};
  int NbrNodes(){return _nbrNodes;};
  iSet * GetSet(){return _pSet;};
  int NbrAssignedSeq(double *q);
  bool FindData(string * motif);
  string * GetError(){return _pError;}

};


#endif
