#ifndef pwm_include_declared
#define pwm_include_declared

#include "inclusive.h"
#include "InstanceMap.h"


class PWM{

 private:
  int _length;
  Matrix _pMatrix;
  double _pPseudo[4];
  double _score;
  double _CSscore;//
  double _ICscore;//
  double _LLscore;//
  double _DLLscore;//
  int _nbrInstances;//
  int _nbrSeqWithInst;//
  string * _id;
  string * _consensus;

  void _ComputeConsensus();

 public: 
  PWM(PWM &pCopy);
  PWM(int W, Matrix pM);
  PWM(int W, Matrix pM, double *snf);
  PWM(int W, Counts pC);
  PWM(int W, Counts pC, double *snf);
  PWM(InstanceMap * pInstances, double *snf, int w);
  ~PWM();

  Matrix GetMatrix() const { return _pMatrix;};
  int Length() const { return _length; };
  string * GetID() const {return _id;};
  double Score() const { return _score;};
  double GetCS() { return _CSscore;};//
  double GetIC() { return _ICscore;};//
  double GetLL() { return _LLscore;};//
  double GetDLL() { return _DLLscore;};//
  int GetNbrInst(){return _nbrInstances;};//
  int GetNbrSeqInst(){return _nbrSeqWithInst;};//
  string * GetConsensus();
  void StderrPrintMatrix();

  double GetValueAt(int i, int j);
  double GetPseudoCountAt(int i);
  char GetConsensusSymbolAt(int index);
  
  double ConsensusScore();
  double InformationContent(double *snf);
  double MaxScore();

  double MutualInformation(PWM *sbjct, int shift);
  double WeightedKullbackLeiberDistance(PWM * sbjct, vector<double>* pVec);

  // adaptors
  PWM * SetID(string * id);
  void SetID(const string &id)
      {if (_id != NULL) delete _id; _id = new string(id); return;};//
  PWM * SetConsensus(string * cons);
  PWM * RebuildMatrix(int W, Counts pC, double *snf);
  PWM * RebuildMatrix(int W, Matrix pM, double *snf);
  PWM * RebuildMatrix(int W, Counts pC);
  PWM * RebuildMatrix(int W, Matrix pM);
  PWM * SetScore(double sc);
  void SetCS(double cs){ _CSscore = cs; return;}; //
  void SetIC(double ic){ _ICscore = ic; return;}; //
  void SetLL(double ll){ _LLscore = ll; return;}; //
  void SetDLL(double dll){ _DLLscore = dll; return;}; //
  void SetNbrInst(int n){ _nbrInstances = n; return;};//
  void SetNbrSeqInst(int n){ _nbrSeqWithInst = n; return;};//
  
  // sub matrix selection
  PWM * SubMatrix(int index, int length);
  PWM * ReverseSubMatrix(int index, int length);
  
  // matrix shuffling
  void ShuffleMatrix();
  PWM * NewShuffledMatrix();
  
};


#endif
