#include "InstanceSet.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2009/07/27
******************************************************************************/
InstanceSet::InstanceSet()
{
  _pSet = new iSet;
  _nbrNodes = 0;
  _nbrSeq = 0;
  _pSeqVolume = NULL;
  _pError = NULL;
}

/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2009/07/27
******************************************************************************/
InstanceSet::~InstanceSet()
{
  if (_pSet != NULL) 
  {
    for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
    {
      // delete instance
      if ((*iter) != NULL)
        delete (*iter);
    }
    // delete set
    delete _pSet;
  }
  _pSet = NULL;

  // _pSeqVolume
  if (_pSeqVolume != NULL)
  {
    for(int n = 0; n < _nbrNodes; n++)
      delete[] _pSeqVolume[n];
    delete[] _pSeqVolume;
  }
  _pSeqVolume = NULL;

  if (_pError != NULL) delete _pError; _pError = NULL; 

}
/******************************************************************************
  Description:  StoreInstance : add to set, merge motif-nbrs from overlappings
  Author_Date:  MC_2009/12/01
******************************************************************************/
void
InstanceSet::StoreInstance(Instance* inst)
{ 
  //cerr << "debug - InstanceSet::StoreInstance - begin" << endl;
  pair<iSetIter, bool> stored;
  // count all the 'loaded instances' (regardless overlapping or exact)
  _nbrNodes++;
  // add new instance in artificial instance per sequence/strand
  // in mode _iStore == 0 - see CompareID() in InstanceSet.h
  stored = _pSet->insert(inst); 
  iSetIter iterFirst = stored.first;

  // if instance from same sequence/strand
  if ( !(stored).second)
  {  
    // create artificial if not yet so
    if ( (*iterFirst)->NbrOcc() == 1)
    { (*iterFirst)->CreateOverlaps();} 
    // add the instance to the big artificial instance
    (*iterFirst)->AddInstance(inst);
  }

  // update the nbr sequences (only if stored, else same seq/strand)
  if ((stored).second) 
  {
    if (_nbrSeq == 0) 
      _nbrSeq++;
    else 
    { // just recount the total nbr sequences 
      _nbrSeq = 1; iter = _pSet->begin();
      iSetIter iter2 = iter; iter2++;
      while (iter2 != _pSet->end())
      {
        if ( (*(*iter)->PrintSeqName()).compare(*(*iter2)->PrintSeqName())) 
        { _nbrSeq++; } // different seqid
        iter++; iter2++;
      }
    }
  }

  //cerr << "Debug-InstanceSet::StoreInstance- end" << endl;
  return;
}
/******************************************************************************
  Description:  NormalizeWeights(const double &norm) 
                of all individiual stored instances
  Author_Date:  MC_2011/09/26
******************************************************************************/
void
InstanceSet::NormalizeWeights(const double &norm)
{
  // at this stage all instances in _pSet are grouped per seq/strand
  // all instances in _pOverlaps from (*iter) are singles 
  vector<Instance*> *pOvlps = NULL;
  Instance * inst = NULL;
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    pOvlps = (*iter)->GetOverlapInsts();
    for (int i = 0; i < (int)pOvlps->size(); i++)
    {
      inst = (*pOvlps)[i];
      inst->NormWeight(norm);
    }
  }
  return;
}

/******************************************************************************
  Description:  RemoveLowProbIndividuals
  Author_Date:  MC_2011/09/27
******************************************************************************/
string
InstanceSet::RemoveLowProbIndividuals(const int &datasize, const double &cut)
{
  string Noisy = "";
  // copy non-rejected instances to new set
  iSet * pNew = new iSet;
  // and keep track to evaluate which sequences are 'noisy' at the end
  vector<string> * pSeqs1 = new vector<string>;

  Instance * inst = NULL;
  double prob; string seq = "";
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    inst = (*iter); 
    seq = *(*iter)->PrintSeqName();
    if ( find(pSeqs1->begin(), pSeqs1->end(), seq) == pSeqs1->end())
    { pSeqs1->push_back(seq);} // store seq-name

    prob = sqrt(inst->GetWeight())*(double)inst->NbrOcc()/datasize;
    if ( prob  < cut)
    {
      // delete node
      delete inst; inst = NULL;
      _nbrNodes--;
      // break : see throw in MotifClusterRun::CheckDatasetDimension()
      if (_nbrNodes < 4 ) break; 
    }
    else
    { 
      inst->SetStoreMode(1);
      pNew->insert(inst);
    }
  }
  // reassign 
  delete _pSet; // only vector
  _pSet = pNew;

  if ((int)_pSet->size() != _nbrNodes)
  { cerr << "--debug::RemoveLowProbIndividuals - error Nodes-size" << endl; }

  if ((int)pSeqs1->size() != _nbrSeq)
  { cerr << "--debug::RemoveLowProbIndividuals - error Seqs1-size" << endl; }

  // recount the nbr of sequences 
  vector<string> * pSeqs2 = new vector<string>;
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    seq = *(*iter)->PrintSeqName();
    if ( find(pSeqs2->begin(), pSeqs2->end(), seq) == pSeqs2->end())
    { pSeqs2->push_back(seq);} // store seq-name
  }
  // reset count
  _nbrSeq = (int)pSeqs2->size();
  // and track the noisy ones
  for (int v = 0; v < (int)pSeqs1->size(); v++)
  {
    seq = (*pSeqs1)[v];
    if ( find(pSeqs2->begin(), pSeqs2->end(), seq) == pSeqs2->end()) 
    {
      if (Noisy != "") Noisy.append(", ");
      Noisy.append(seq);
    }
  }
  // cleanup
  delete pSeqs1; pSeqs1 = NULL;
  delete pSeqs2; pSeqs2 = NULL;	
  return Noisy;
}

/******************************************************************************
  Description:  BuildAsymmetricalMatrix : _C1*N
      M(i,j) = set i has instance j, 0/1 or a probability value
  Author_Date:  MC_2009/08/06
******************************************************************************/
double * 
InstanceSet::BuildAsymmetricalMatrix(vector<string*> * pAllSets)
{
  int C1 = (int)pAllSets->size();
  bool bProceed = 1;
  if (_pError != NULL) delete _pError; _pError = NULL;
  // variable to return
  double * asymmetric = new double[C1*_nbrNodes];
  // set all elements to 0
  for(int i = 0; i < C1*_nbrNodes; i++)
    asymmetric[i] = 0;

  // check for each node contents of _pJoinMotifIDs and _pJoinWeights
  vector<string> * pJoinData = NULL;
  vector<double> * pJoinWs = NULL;
  double weight;
  int i;
  iter = _pSet->begin(); 
  for(int j = 0; j < _nbrNodes; j++)  // instance kolom j
  { 
    pJoinWs = (*iter)->GetAllWeights();
    pJoinData = (*iter)->GetAllMotifs();
    // iterate through all 'data' of this instance
    for(int m = 0; m < (int)pJoinData->size(); m++)
    {
      weight = (*pJoinWs)[m]; // get corresponding weight
      // find position of this data-id in pAllSets
      i = 0;
      for(; i < (int)pAllSets->size(); i++)
      { 
        if ( (*pJoinData)[m] == *(*pAllSets)[i]) 
          break;
      }
      // store "1" or "weight" at T(i,j) intersection
      // dit veronderstelt wel dat je geen overlappende instanties 
      // in eenzelfde data-input mag hebben !!
      // als dat toch zo zou zijn, geef warning
      if ( asymmetric[j*C1+i] != 0)
      { // keep double reporting (to screen and to file, even if GFFWriter would be screen
        // reason : to get reporting via MotifSuite webserver-use
        cerr << "WARNING:: BuildAsymmetricalMatrix() : motif >" << (*pJoinData)[m]
        << " describes the same (largely overlapping) instance twice : ";
        string *id = (*iter)->InstanceID(); cerr << *id; 
        cerr << ". Only the highest-weight instance is further considered. You may want to re-evaluate your inputfile. " << endl;
        // store for transfer to report
        ostringstream cerrstr;
        cerrstr << "WARNING:: BuildAsymmetricalMatrix() : motif >" << (*pJoinData)[m]
        << " describes the same (largely overlapping) instance twice : " << *id; 
        delete id; id = NULL;
        cerrstr << ". Only the highest-weight instance is further considered. You may want to re-evaluate your inputfile. " << endl;
        _pError = new string(cerrstr.str()); // will be uploaded later on
        cerrstr.flush();
        // no error, only warning, keep the highest one
        bProceed = 1; // recode to 0 in case abortion preferred
        if (asymmetric[j*C1+i] < weight){asymmetric[j*C1+i] = weight;}
      }
      else
      { asymmetric[j*C1+i] = weight; } 
    }
    // go to next node
    iter++;
  }
  if (!bProceed)
  { delete[]asymmetric; asymmetric = NULL;}

  return asymmetric;
}
/******************************************************************************
  Description:  AssignNodesToSeqs() : identify to which sequences each instance belongs
    // given that instances are stored in grouped seqid-order in InstanceSet !
  Author_Date:  MC_2011/09/28
******************************************************************************/
void 
InstanceSet::AssignNodesToSeqs()
{
  //cerr << "InstanceSet::AssignNodesToSeqs()-begin" << endl;

  // variable to store locally _nbrNodes * _nbrSeq
  _pSeqVolume = new double*[_nbrNodes];
  for(int n = 0; n < _nbrNodes; n++)
  {
    _pSeqVolume[n] = new double[_nbrSeq];
    for(int j = 0; j < _nbrSeq; j++)
      _pSeqVolume[n][j] = 0;
  }

  // first node belongs to first sequence
  _pSeqVolume[0][0] = 1;

  int i = 1; int j = 0;
  iter = _pSet->begin(); iter++; // second node
  iSetIter iter0 = _pSet->begin(); // compare with earlier id
  for(; iter != _pSet->end(); iter++)
  { 
    if ( (*(*iter)->PrintSeqName()).compare(*(*iter0)->PrintSeqName()) ) 
    {  j++; } // new sequence
    // assign 1 at instance-sequence intersection
    _pSeqVolume[i][j] = 1;
    iter0++; 
    i++;
  }
  //cerr << "InstanceSet::AssignNodesToSeqs()-end" << endl;
  return;
}

/******************************************************************************
  Description:  NbrAssignedSeq(q) : count nbr of seq spanned by cluster Q
	  for reporting purposes
	  assumes q is of length _nbrNodes (N)
  Author_Date:  MC_2009/07/26
******************************************************************************/
int
InstanceSet::NbrAssignedSeq(double * q)
{  
  int nbr(0);
  double n;
  // compute matrixproduct of q * _seqVolume
  for(int j = 0; j < _nbrSeq; j++)
  {
    n = 0;
    for(int i = 0; i < _nbrNodes; i++)
      n += q[i] * _pSeqVolume[i][j];
    if (n != 0) nbr++;
  }
  return nbr;
}
/******************************************************************************
  Description:  GroupExactInstances() : merge the exact occurences together
  Author_Date:  MC_2009/12/11
******************************************************************************/
void
InstanceSet::GroupExactInstances()
{
  //cerr << "debug - InstanceSet::GroupExactInstances - begin" << endl;
  vector<Instance*> * pArtificials(NULL), *pArtificialsCopy(NULL); 
  Instance * pInst1(NULL), *pInst2(NULL);
  bool same;

  // for each artificial instance 
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    if ( (*iter)->GetOverlapInsts() != NULL) // artifical trigger
    {
      // get the set of instances part of this artificial
      pArtificials = (*iter)->GetOverlapInsts();
      // create a copy vector where exacts will be melted
      pArtificialsCopy = new vector<Instance*>;
      // sequentially process all instances from pArtificials
      for(int i = 0; i < (int)pArtificials->size(); i++)
      { 
        pInst2 = (*pArtificials)[i]; // a single instance
        same = false;
        // check if an exact occurence was already kept 
        for(int j = 0; j < (int)pArtificialsCopy->size(); j++)
        {
          pInst1 = (*pArtificialsCopy)[j];
          if ( pInst1->Start() == pInst2->Start() && pInst1->Length() == pInst2->Length())
          { 
            // create artificial if not yet so
            if ( pInst1->NbrOcc() == 1)
            { pInst1->CreateOverlaps();} 
            // add pInst2 to the 'exacts' artificial instance
            pInst1->AddInstance(pInst2);
            same = true; 
            // update the _nbrInst
            _nbrNodes--;
            break;
          }
          pInst1 = NULL;
        }
        if (!same) // add pInst2 to copy
        { pArtificialsCopy->push_back(pInst2);}
        pInst2 = NULL;
      }   
      // leave a group of fully exacts for what it is now 
      delete pArtificials; // this deletes the vector of _pOverlaps
      (*iter)->SetOverlaps(pArtificialsCopy); // assign new vector
      pArtificialsCopy = NULL;
      pArtificials = NULL;
    }
  }
  //cerr << "debug - InstanceSet::GroupExactInstances - end" << endl;
  return;
}
/******************************************************************************
  Description:  SplitOverlappingInstances() : split up artificials in subsets of
                highest occ/weight/length artificials and melt lower when
                there is min50% overlap
  Author_Date:  MC_2009/12/12
******************************************************************************/
void
InstanceSet::SplitOverlappingInstances()
{
  //cerr << "debug - InstanceSet::SplitOverlappingInstances - begin" << endl;
  vector<Instance*> * pArtificials(NULL), *pArtificialsCopy(NULL); 
  Instance * pInst1(NULL), *pInst2(NULL);
  vector<int> * pMask = NULL;
  vector<int> * pOvl = NULL;
  vector<int> * pTrack = NULL;
  bool done;
  int Imax,maxocc,length,minovl,costart,coend,ovl;
  double weight;

  // make a new set where all will be stored in mode==1 at the very end
  iSet * pNewSet = new iSet;

  // for each instance of the existing set
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    if ( (*iter)->PrintSite() != NULL )
    { // a single instance
      (*iter)->SetStoreMode(1);
      pNewSet->insert(*iter); 
    }
    else // may also be a fully group of exacts
    { 
      //// artificial instance first needs a splitup of overlapping
      //-------------------------------------------------------------
      // get the set of instances part of this artificial
      pArtificials = (*iter)->GetOverlapInsts();
      pMask = new vector<int>((int)pArtificials->size(), 1);
      pOvl = new vector<int>((int)pArtificials->size(), 0);
      pTrack = new vector<int>((int)pArtificials->size(), -1);
      // create a copy vector where splits will be stored & overlaps melted
      pArtificialsCopy = new vector<Instance*>;
      //// sequentially process all 'main' instances from pArtificials
      done = 0;
      while(!done)
      {
        // in order from the highest occ a/o weight a/o length instance
        Imax = -1; maxocc = -1; length = -1; weight = -1; 
        for(int i = 0; i < (int)pArtificials->size(); i++)
        {
          if ( (*pMask)[i] == 1) // non-overlap high candidate
          { pInst1 = (*pArtificials)[i];
            if (pInst1->NbrOcc() > maxocc || 
                (pInst1->NbrOcc() == maxocc && 
                (pInst1->GetWeight() > weight || 
                (pInst1->GetWeight() == weight && pInst1->Length() > length))))
            {
              maxocc = pInst1->NbrOcc(); length = pInst1->Length(); 
              weight = pInst1->GetWeight(); Imax = i; //already avgd weight
            }
          }
        }
        if (Imax != -1)
        {
          // pInst1 is the best non-ovl instance at this point : store as main
          pInst1 = (*pArtificials)[Imax];
          pArtificialsCopy->push_back(pInst1);
          (*pMask)[Imax] = 0;
          // compute for all other non-masked instances the overlap with pInst1
          for(int j = 0; j < (int)pArtificials->size(); j++)
          {
            if ((*pMask)[j] == 0) continue;
            pInst2 = (*pArtificials)[j];
            // minimal overlap = 1/2 of shortest instance
            minovl = int(pInst1->Length()/2);
            if (pInst1->Length() > pInst2->Length()){minovl = int(pInst2->Length()/2);}
            // compute overlap and track maximal overlap-pointer
            if (pInst1->Start() > pInst2->Start()){costart = pInst1->Start();}
            else { costart = pInst2->Start();} 
            if (pInst1->Start() + pInst1->Length() < pInst2->Start() + pInst2->Length())
            {coend = pInst1->Start() + pInst1->Length();} 
            else {coend = pInst2->Start() + pInst2->Length();}
            ovl = coend - costart; 
            if (ovl < 0 || ovl < minovl) { continue;}
            if ( ovl > (*pOvl)[j])// check if higher than earlier assigned
            { 
              (*pOvl)[j] = ovl;
              (*pTrack)[j] = Imax;
              (*pMask)[j] = -1;
            }
          }
        }
        // trigger end of storing mains 
        done = 1; // assume
        for(int i = 0; i < (int)pArtificials->size(); i++)
        { if ( (*pMask)[i] == 1){ done = 0; break;}}
      }
      // for each main-instance, add the overlapping instances
      for(int i = 0; i < (int)pArtificialsCopy->size(); i++)
      {
        pInst1 = (*pArtificialsCopy)[i];
        // create artificial by default (also if nothing would be added after all)
        if ( pInst1->NbrOcc() == 1)
        { pInst1->CreateOverlaps();} 
        // search pointer in pArtificials
        int j = 0;
        for( ;j < (int)pArtificials->size(); j++)
        { if ( pInst1 == (*pArtificials)[j])
          { break;}
        } // j is the instance to be tracked
        // get all the overlapping instances that track with position j
        for(int jj = 0; jj < (int)pArtificials->size(); jj++)
        {
          if ( (*pMask)[jj] == -1 && (*pTrack)[jj] == j) 
          { 
            // add this instance to the artificial instance
            // exacts are split up again in individual addings
            pInst1->AddInstance((*pArtificials)[jj]);
            (*pMask)[jj] = 0;
          }
        }
      }
      //// now add all items from pArtificialsCopy to new set
      for(int i = 0; i < (int)pArtificialsCopy->size(); i++)
      {
        // first make them non-artificial
        (*pArtificialsCopy)[i]->UpdateInstance();
        (*pArtificialsCopy)[i]->SetStoreMode(1);
        pNewSet->insert((*pArtificialsCopy)[i]);
      }
      // cleanup
      delete pArtificials; pArtificials = NULL;
      (*iter)->SetOverlaps(NULL); delete (*iter);
      delete pArtificialsCopy; pArtificialsCopy = NULL;
      delete pMask; pMask = NULL;
      delete pOvl; pOvl = NULL;
      delete pTrack; pTrack = NULL;
    }
  }
  // now delete the existing set (not contents)
  delete _pSet;
  // and assign the new split set to it
  _pSet = pNewSet;
  _nbrNodes = (int)pNewSet->size();

  return;
}
/******************************************************************************
  Description:  FindData(string * motif) : check if any of the instances was
	  detected by the given motif
  Author_Date:  MC_2011/09/27
******************************************************************************/
bool
InstanceSet::FindData(string * motif)
{  
  bool found = 0;
  vector<string> * pGetMotifs = NULL;
  // scan through all instances (may be groups)
  for(iter = _pSet->begin(); iter != _pSet->end(); iter++)
  {
    // scan the set of stored 'data' for this instance
    pGetMotifs = (*iter)->GetAllMotifs();
    for(int m=0; m < (int)pGetMotifs->size(); m++)
    {
      if ( *motif == (*pGetMotifs)[m])
      { found = true; break;}
    }
    if (found) break;
  }
  return found;
}
