#include "MotifClusterRun.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2009/07/24
******************************************************************************/
MotifClusterRun::MotifClusterRun()
{
  _pAllInstances = new InstanceSet();
  _C1 = 0; 
  _pOutText = NULL;
  _pDebugText = NULL;
  _pClusterResults = new list<Cluster*>;
  _nbrClusters = 0;
  _pAllData = NULL;
}

/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2009/07/24
******************************************************************************/
MotifClusterRun::~MotifClusterRun()
{
  _pOutText = NULL; _pDebugText = NULL; // in main

  // _pAllInstances and contents
  if (_pAllInstances != NULL)
    delete _pAllInstances;
  _pAllInstances = NULL;   

  // _pClusterResults and contents
  list<Cluster*>::iterator iter = _pClusterResults->begin();
  if (_pClusterResults != NULL)
  {
    for(; iter != _pClusterResults->end(); iter++)
    {
      if ((*iter) != NULL) // delete Cluster-object
        delete (*iter);
      (*iter) = NULL;
    }
    delete _pClusterResults;// delete list
  }
  _pClusterResults = NULL;

  // _pAllData and contents
  if (_pAllData != NULL)
  {
    for(int i = 0; i < (int)_pAllData->size();i++)
    { if ((*_pAllData)[i] != NULL) 
        delete (*_pAllData)[i]; 
      (*_pAllData)[i] = NULL;
    }
    delete _pAllData;
  }
  _pAllData = NULL;

}
/******************************************************************************
  Description:  LoadInstances into _pAllInstances and _pAllData
                format '>motif or #id
                        instances
  Author_Date:  MC_2011/09/22
******************************************************************************/
void
MotifClusterRun::LoadInstances(string * file)
{
  //cerr << "debug - MotifClusterRun::LoadInstances - begin" << endl;
  // open file and set cursor on first > or #id
  FastaIO *fileIO = new FastaIO(file);
  // trace inputSets
  _pAllData = new vector<string*>;
  bool empty;
  double normWeight = 1; // keep track of maxWeight to normalize

  // in case of MS-3.1.5 .txt file, lines start with #id: ....
  string tab("\t");
  string meaningful("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890");
  string::size_type pos = 0;
  string start("#");
  
  try
  {
    // check file validity
    if (!fileIO->IsOpen()) { delete fileIO; fileIO = NULL;
      throw ErrorInputfiles((*file).append(" not available"), _pOutText);}
    if (!fileIO->HasNext()){ delete fileIO; fileIO = NULL;
      throw ErrorInputfiles((*file).append(" has no > or #id lines"), _pOutText);}
    string motif = "unknown";
    // read input from file  
    while ( fileIO->IsOpen() && fileIO->HasNext() ) 
    {
      // line starts with > or #id: 
      motif = fileIO->ReadLine();
      // skip lines with >>
      if( motif[1] == '>' ) continue;
      // read motifid, if not assign by _C1 motifcounter-number
      _C1++;       
      if ( motif[0] == '>') motif.erase(0,1); // erase leading >
      else // erase leading #id: and only keep first id (untill tab)
      { motif.erase(0,3); 
        if ( motif[0] == ':') motif.erase(0,1);
        if ( motif[0] == ' ') motif.erase(0,1);
        if ((pos = motif.find_first_of(tab)) != string::npos)
          motif = motif.substr(0, pos-1);
      }

      if ((pos = motif.find_first_of(meaningful)) == string::npos) 
      {
        ostringstream number; number << "motif_" << _C1; 
        motif = number.str(); number.flush();
      }

      // store inputsetid
      // check for double entries
      int i = 0;
      for(; i < (int)_pAllData->size(); i++)
      { 
        if ( *(*_pAllData)[i] == motif) 
          break;
      }
      if ( i == 0 || i == (int)_pAllData->size())
      { _pAllData->push_back(new string(motif));}
      else
      {
        throw InvalidInput(motif.append(" is described twice"), _pOutText);
      }
      // at this point no instances assigned yet
      empty = true;
      // read the following lines with instances
      while( fileIO->IsOpen() ) 
      { 
        // break at next motif
        if (fileIO->HasNext()) { break;} // break occurs at next > or #id line
		// but do not read other # lines (this is not > or not #id lines)
        if (fileIO->CheckLineStart(start)){continue;}

        // read instance
        Instance * pInst = fileIO->ReadInstance(motif);
        if (pInst == NULL) 
          throw InvalidInput(motif.append(" incorrect instance format"), _pOutText);
        // track the instance weight
        if (pInst->GetWeight() > normWeight){normWeight = pInst->GetWeight();}
        // the instances are stored in big groups ('artificial instances')
        // of (possibly only partially) overlapping instances 
        _pAllInstances->StoreInstance(pInst); pInst = NULL;
        empty = false;
      }
      // skip inputSet without instances
      if (empty && _C1 > 0) 
      { _C1--; _pAllData->pop_back();}
    }
  }
  catch (InvalidInput)
  { delete fileIO; fileIO = NULL; throw ErrorInputfiles("in -f file", _pOutText);}

  // close instancesfile
  delete fileIO;

  // now normalize all the loaded weights with normWeight
  if (normWeight != 1) {_pAllInstances->NormalizeWeights(normWeight);}

  ostringstream report; // to debugfile
  bool bReport = (_pDebugText != NULL);
  iSet * pSet; iSetIter iter;
  if (bReport)
  {
    report << "debug LoadInstances() print stored instances" << endl;
    if (normWeight != 1) report << "all weights have been normalized with " << normWeight << endl; 
    pSet = _pAllInstances->GetSet();
    iter = pSet->begin();
    for(; iter != pSet->end(); iter++)
    {
      string * id = (*iter)->InstanceID();
	  if (1)//id->substr(4,1) == "0")
      { report << *id << "(" << (*iter)->NbrOcc() 
                      << "," << setprecision(2) << (*iter)->GetWeight() << ")"; // per sequence/strand
        delete id; id = NULL;
        vector<Instance*>* pOvlps = (*iter)->GetOverlapInsts();
        if (pOvlps != NULL)
        {
          report <<  ":[";
          for(int i = 0; i < (int)pOvlps->size(); i++)
          {
            id = (*pOvlps)[i]->InstanceID();
            report <<  *id << "(" << (*pOvlps)[i]->NbrOcc() 
                           << "," << setprecision(2) <<(*pOvlps)[i]->GetWeight() << "),";
            delete id; id = NULL;
          }
          report <<  "]" << endl;
        } 
        else report << "[no]" << endl;
      }
      if (id != NULL) delete id; id = NULL;
    } 
    report << "debug LoadInstances() : size_stored = " << (int)pSet->size() << endl;
  }

  // group exact occurences into artificals together
  _pAllInstances->GroupExactInstances();

  if (bReport)
  {
    report << "debug LoadInstances() print grouped instances" << endl;
    pSet = _pAllInstances->GetSet();
    iter = pSet->begin();
    for(; iter != pSet->end(); iter++) 
    {
      string * id = (*iter)->InstanceID();
	  if (1)//id->substr(4,1) == "0")
      { report << *id << "(" << (*iter)->NbrOcc() 
                      << "," << setprecision(2) <<(*iter)->GetWeight() << ")"; // per sequence/strand
        delete id; id = NULL;
        vector<Instance*>* pOvlps = (*iter)->GetOverlapInsts();
        vector<Instance*>*pExacts = NULL;
        if (pOvlps != NULL)
        {
          report <<  ":[";
          for(int i = 0; i < (int)pOvlps->size(); i++)
          {
            pExacts = (*pOvlps)[i]->GetOverlapInsts();
            if (pExacts == NULL)
            {id = (*pOvlps)[i]->InstanceID();}
            else { id = (*pExacts)[0]->InstanceID();}
            report <<  *id << "(" << (*pOvlps)[i]->NbrOcc() 
                           << "," << setprecision(2) <<(*pOvlps)[i]->GetWeight() << "),";
            delete id; id = NULL;
          }
          report <<  "]" << endl;
        } 
        else report << "[no]" << endl;
      }
      if (id != NULL) delete id; id = NULL;
    } 
    report << "debug LoadInstances() : size_grouped = " << (int)pSet->size() << endl;
  }

  // splitup each artifical in a non-overlapping subset of representatives
  _pAllInstances->SplitOverlappingInstances();

  if (bReport)
  {
    report << "debug LoadInstances() print split nodes" << endl;
    pSet = _pAllInstances->GetSet();
    iter = pSet->begin();
    for(; iter != pSet->end(); iter++)
    {
      string * id = (*iter)->InstanceID();
	  if (1) //id->substr(4,1) == "0")
      { report << *id << "(" << (*iter)->NbrOcc()
                      << "," << setprecision(2) <<(*iter)->GetWeight() << ")"; // per node
        delete id; id = NULL;
        id = (*iter)->GetOverlaps();
        if ( id != NULL)
        { report <<  ":[" << *id << "]" << endl;
          delete id; id = NULL;
        } else report << endl;
      }
      if (id != NULL) delete id; id = NULL;
    } 
    report << "debug LoadInstances() : size_split = " << (int)pSet->size() << endl;
    // add to file/screen
    _pDebugText->AddComment(report.str());
    report.flush(); // flush
  }
  //cerr << "debug - MotifClusterRun::LoadInstances - end" << endl;
  return; 
}

/******************************************************************************
  Description:  DatasetDimension() : if not ok, exit to main
  Author_Date:  MC_2011/09/22
******************************************************************************/
void 
MotifClusterRun::DatasetDimension(const double &pCut, const double &iCut,
                                  const double &jCut)
{
  //cerr << "debug - MotifClusterRun::CheckDatasetDimension - begin" << endl;
  // report some results
  ostringstream out_1;
  out_1 << "#DatasetDimension LOADED : " << "nbr instances = " 
       << _pAllInstances->NbrNodes() << ", nbr sequences = " 
       << _pAllInstances->NbrSeq() << ", nbr motifs = " << _C1 << endl;
  int noise1 = _pAllInstances->NbrSeq();

  if (_C1 < 2) 
    throw InputDimension("inputfile : at least 2 motifs required", _pOutText);
    // catch in main
  if (_pAllInstances->NbrNodes() < 4) 
    throw InputDimension("inputfile : at least 2 instances required", _pOutText);
    // catch in main
  if (_pAllInstances->NbrSeq() < 2) 
    throw InputDimension("inputfile : at least 2 sequences required", _pOutText);
    // catch in main

  // first set the _minSeqVolume
  double treshold = _pAllInstances->NbrSeq() * iCut;
  if ((treshold - int(treshold)) >= 0.5){ _minSeqVolume = int(treshold) +1;}
  else { _minSeqVolume = int(treshold);}   
  // and the _minDataVolume
  treshold = _C1 * jCut;
  if ((treshold - int(treshold)) >= 0.5){ _minDataVolume = int(treshold) +1;}
  else { _minDataVolume = int(treshold);} 


  // now remove instances with a very low detection frequency in data
  // i.e. freq*sqrt(w) < MinNodeFreq
  string noisyseq = _pAllInstances->RemoveLowProbIndividuals(_C1, pCut);
  noise1 -= _pAllInstances->NbrSeq();
  // DO NOT reset _minSeqVolume although NbrSeq() might have changed now
  // and also construct _pSeqVolume matrix (node-seq assignment)
  _pAllInstances->AssignNodesToSeqs();
  
  // remove noisy data from _pAllData 
  // (where all instances have been filtered out)
  int noise2(0);
  string noisydata = "";
  vector<string*> *pCopy = new vector<string*>;
  string * pMotif = NULL; 
  for (int d = 0; d < (int)_pAllData->size(); d++)
  {
    pMotif = (*_pAllData)[d];
    // see if the 'data' is stored in an instance
    if ( _pAllInstances->FindData(pMotif))
    { pCopy-> push_back(pMotif);}
    else  
    { 
      if (noisydata != "") noisydata.append(", ");
      noisydata.append(*pMotif);
      noise2++;
      delete pMotif; // delete string
    } 
  }
  delete _pAllData; // delete vector
  _pAllData = pCopy;
  // DO NOT reset _minDataVolume although #data might have changed now
  
  if (_pAllInstances->NbrNodes() < 4) 
    throw InputDimension("inputfile : not enough instances with sufficient frequency (-p) - clustering not applicable", _pOutText);
    // catch in main
  if (_pAllInstances->NbrSeq() < _minSeqVolume) 
    throw InputDimension("inputfile : not enough sequences (-i) have an instance with sufficient frequency (-p) - clustering not applicable", _pOutText);
  if ((int)_pAllData->size() < _minDataVolume) 
    throw InputDimension("inputfile : not enough motifs (-j) have an instance with sufficient frequency (-p) - clustering not applicable", _pOutText);
  // catch in main
  if (_pAllInstances->NbrNodes() > 10000) 
     throw InputDimension("inputfile : too much instances - abort for computational reasons", _pOutText);
    // catch in main

  out_1 << "#DatasetDimension FILTERED (parameter -p) : nbr instances = " 
       << _pAllInstances->NbrNodes() << ", nbr sequences = " 
       << _pAllInstances->NbrSeq() << ", nbr motifs = " << (int)_pAllData->size()  << endl;
  if (noise1 > 0)
  { out_1 << "#::Noisy sequences(" << noise1 << "):[" << noisyseq << "]" << endl;}
  if (noise2 > 0)
  { out_1 << "#::Noisy motifs(" << noise2 << "):[" << noisydata << "]" << endl;}
  out_1 << "#List of merged non-spurious instances :" << endl;

  // print all to out_1
  iSet * pSet = _pAllInstances->GetSet();
  iSetIter iter = pSet->begin();
  for(; iter != pSet->end(); iter++)
  {
    string * id = (*iter)->InstanceID();
    out_1 << *id << "(" << (*iter)->NbrOcc() << ")"; 
    delete id; id = NULL;
    id = (*iter)->GetOverlaps();
    if ( id != NULL)
    { out_1 <<  ":[" << *id << "]" << endl;
      delete id; id = NULL;
    } 
    else out_1 << endl;
  } 

  // report on screen or file
  _pOutText->AddComment(out_1.str());
  out_1.flush();
  return;
}

/******************************************************************************
  Description:  ClusterXnodesYdata() : 
                relies on edgeclust.m function in libClust
  Author_Date:  MC_2011/09/22
******************************************************************************/
bool 
MotifClusterRun::ClusterXnodesYdata(const double &mCut,const double &minCS)
{
  int N = _pAllInstances->NbrNodes();
  int O = (int)_pAllData->size(); // number of 'data'

  // report some results
  ostringstream out_1;

  // build asymmetrical M(i,j)=data i has instance j (MxN) with weight 
  double * matrixFasym = _pAllInstances->BuildAsymmetricalMatrix(_pAllData);

  if (_pAllInstances->GetError() != NULL)
  {
    out_1 << *(_pAllInstances->GetError()) << endl;
    // report on screen or file
    _pOutText->AddComment(out_1.str());
    out_1.flush();
  }
  if (matrixFasym == NULL) // some error in input
  {
    out_1 << "#There are errors in your inputfile. Please contact us." << endl;
    // report on screen or file
    _pOutText->AddComment(out_1.str());
    out_1.flush();
    return 0;
  }
  else 
  { if(_pDebugText != NULL) 
    { 
      ostringstream out_2;
      out_2 << "#matrix F - ready for clustering " << endl;
      out_2 << "#instances:[";
      iSet * pSet = _pAllInstances->GetSet();
      iSetIter iter;
      for(iter = pSet->begin(); iter != pSet->end(); iter++)
      {
        string *id = (*iter)->InstanceID();
        out_2 << *id << "\t";
        delete id; id = NULL;
      }
      for (int i=0; i < O; i++)
      {
        out_2 << *(*_pAllData)[i] << "\t[";
        for(int j=0; j < N; j++)
        { out_2 << matrixFasym[j*O+i] << "\t";}
        out_2 << "]" << endl;
      }
      _pDebugText->AddComment(out_2.str());
      out_2.flush();
    }
  }
  mwArray Fasym(O, N, mxDOUBLE_CLASS);
  Fasym.SetData(matrixFasym,O*N);
  delete[] matrixFasym; matrixFasym = NULL;

  // execute the clustering in matlab
  mwArray nodes; // memberships of nodes
  mwArray data; // memberships of data
  mwArray excludes; // number of excluded elements per cluster in nodes/data
  mEdgeClust(3, data, nodes, excludes, Fasym, mwArray(mCut));

  // store the results
  int nbrClusters = (int)nodes.GetDimensions()(1,2); // # columns 
  out_1 << "#MotifClusterRun::Extraction(): " << nbrClusters<<" clusters extracted by "
        << "edgeclust.m " << endl;

  Cluster * pResult = NULL;
  int reject_i = 0; // low sequence volume
  int reject_j = 0; // low data volume
  int reject_c = 0; // low cs score

  // repeat for each cluster
  for(int c = 1; c <= nbrClusters; c++)
  {
    // convert nodes mwArray to double*
    double * pCluster = new double[N];
    for(int j = 1; j <= N; j++) pCluster[j-1] = nodes(j,c);
    // check low seqvolume
    if (_pAllInstances->NbrAssignedSeq(pCluster) <= _minSeqVolume)
    { delete pCluster; pCluster = NULL;
      reject_i++;
      continue;
    }
    // convert data mwArray to double*
    double * pdata = new double[O];
    int nnz = 0;
    for(int j = 1; j <= O; j++) 
    { pdata[j-1] = data(j,c); 
      if (pdata[j-1] > 0) nnz++;
    }
    // check low datavolume
    if (nnz <= _minDataVolume)
    { delete pCluster; pCluster = NULL;
      delete pdata; pdata = NULL;
      reject_j++;
      continue;
    }
    // create an object 
    pResult = new Cluster(N, pCluster, (int)excludes(1,c), 
						  O, pdata, (int)excludes(2,c));
    // reconstruct PWM in Cluster-object
    pResult->ReconstructMotif(_pAllInstances);
    // check on minimal consensus-score
    if (pResult->GetCSMotif()->GetCS() <= minCS)
    { //delete pCluster; pCluster = NULL; // done in ~pResult
      //delete pdata; pdata = NULL;// done in ~pResult
      delete pResult; pResult = NULL;
      reject_c++;
      continue;
    }
    // store
    _pClusterResults->push_back(pResult);
    _nbrClusters++;
  }

  // report
  out_1 << "# " << reject_i << " clusters rejected for low (<= " << _minSeqVolume
        << ") number of sequences (threshold derived from parameter -i)" << endl;
  out_1 << "# " << reject_j << " clusters rejected for low (<= " << _minDataVolume
        << ") number of corresponding motifs in input file (threshold derived from parameter -j)" << endl;
  out_1 << "# " << reject_c << " clusters rejected for low (<= " << minCS
        << ") Consensus score (threshold set by parameter -c)" << endl;

  // report on screen or file
  _pOutText->AddComment(out_1.str());
  out_1.flush();
   
  return 1;
}

/******************************************************************************
  Description:  PrintClusters : to _pOutText
                straightforward print of Cluster* 
  Author_Date:  MC_2009/09/15
******************************************************************************/
void 
MotifClusterRun::PrintClusters()
{
  // send text to ostringstream
  ostringstream text;
  text << "#PrintClusters:: " << _nbrClusters << " ensemble motifs :" << endl;

  // iterate through all clusters
  int i = 1; double member; double value;
	//some variables
  double * pClustNodes(NULL), *pClustShifts(NULL), *pClustData(NULL);
  string * pClustCSdist(NULL);
	
  PWM * pMotifCS = NULL; PWM* pMotifW = NULL;
	
  iSet * pSet = _pAllInstances->GetSet();
  iSetIter iter; int j;
	
  string * id = NULL;
	
  list<Cluster*>::iterator iterC = _pClusterResults->begin();
  for(; iterC != _pClusterResults->end(); iterC++)
  {
    // compose id-info 
    ostringstream clusterid;
    clusterid << "box_" << i;
    pMotifCS = (*iterC)->GetCSMotif();
    clusterid << "_CS[" << pMotifCS->Length() << "," 
              << *pMotifCS->GetConsensus() 
               << "," << setprecision(3) << pMotifCS->GetCS() << "]";
    pMotifW = (*iterC)->GetWidthMotif();
    if (pMotifW != NULL) // one after the other
    { clusterid << "_W[" << pMotifW->Length() << "," 
                << *pMotifW->GetConsensus() 
                << "," << setprecision(3) << pMotifW->GetCS() << "]";
    }
    value = (*iterC)->NodeThresh();
    if (value < 0.01) clusterid << "_%InstanceCut<0.01";
    else clusterid << "_%InstanceCut=" << setprecision(2) << value;
    value = (*iterC)->DataThresh();
	if (value < 0.01) clusterid << "_%MotifCut<0.01";
    else clusterid << "_%MotifCut=" << setprecision(2) << value;
	  
	  // added 06 april 2011 : report nbr seq with at least one instance
    value = (double)_pAllInstances->NbrAssignedSeq((*iterC)->GetClusterNodes());
    clusterid << "_nbrSeq=" << (int)value;
    clusterid << "_nbrInst=" << (*iterC)->NbrNNZNodes();
    clusterid << "_nbrMotifs=" << (*iterC)->NbrNNZData();

    // add this clusterid to PWM-object and 'text'
    if (pMotifW == NULL) 
    { pMotifCS->SetID(clusterid.str());
    }
    else 
    {
      ostringstream idcs; idcs << clusterid.str() << "_[PWM=CS-based]";
      pMotifCS->SetID(idcs.str()); idcs.flush(); 
      ostringstream idw; idw << clusterid.str() << "_[PWM=width-based]";
      pMotifW->SetID(idw.str()); idw.flush();
    }
    text << endl << "#id: " << clusterid.str()<< endl;
    pMotifCS = NULL; pMotifW = NULL;
    clusterid.flush();
	  
	  
    // print CS-distribution
    pClustCSdist = (*iterC)->GetCSdistribution();
    if (pClustCSdist!= NULL)
    { text << "#CS-distribution : " << *pClustCSdist << endl;}
	  

    // iterate through _pAllInstances to report positive nodes
    pClustNodes = (*iterC)->GetClusterNodes();
    pClustShifts = (*iterC)->GetClusterShifts();
    // compute weighted occurencies
	double weightedocc = 0;
	double sumW = 0;
    j = 0;
    for(iter = pSet->begin(); iter != pSet->end(); iter++)
    {
      // sum (product of membership and occ)
      weightedocc += ( pClustNodes[j] * (*iter)->NbrOcc());
      // sum memberships
      sumW += pClustNodes[j];
      j++; // next node
    }
    weightedocc /= sumW; // normalize
    int occW = -1;
    if (weightedocc - int(weightedocc) < 0.5){occW = (int)weightedocc;}
    else {occW = int(weightedocc) +1;}  
	  
    //report node memberships
    text << "#instance memberships: " 
         << "( N(i)=" <<(*iterC)->WeightedInstanceCount()<< ", "
         << "Occ(i)=" << occW << ")" << endl;
    j = 0;
    for(iter = pSet->begin(); iter != pSet->end(); iter++)
    {
      member = pClustNodes[j];
      if ( member != 0) 
      { 
        id = (*iter)->InstanceID(); //sequence-id_start_end_strand_site
        text << *id; 
        text << "_(shift=" << pClustShifts[j];
        text << ",occ=" << (*iter)->NbrOcc() << ")";
        text << "\t" << setprecision(3) << member << endl;
        delete id; id = NULL;
      } 
      j++; // next node
    }

    pClustData = (*iterC)->GetClusterData();
    text << "#motif memberships: ( N(m)=" 
         << (*iterC)->WeightedMotifCount() << ")" << endl;
    for(int m = 0; m < (int)_pAllData->size(); m++)
    {
      member = pClustData[m];
      if (member != 0)
      { 
        text << *(*_pAllData)[m]; // data-id
        text << "\t" << setprecision(3) << member << endl;
      }
    }

    // augment index
    i++;
  }
  // add to file/screen
  _pOutText->AddComment(text.str());
  text.flush(); // flush
  pClustNodes = NULL; pClustShifts = NULL; 
  pMotifCS = NULL; pMotifW = NULL; pClustData = NULL;
  return;
}
/******************************************************************************
  Description:  PrintMotifs in pwm format to -O file
  Author_Date:  MC_2009/07/27
******************************************************************************/
void 
MotifClusterRun::PrintMotifs(PWMIO * matrixIO)
{
  PWM * pMotif = NULL;
  list<Cluster*>::iterator iterC = _pClusterResults->begin();
  for(; iterC != _pClusterResults->end(); iterC++)
  {
    pMotif = (*iterC)->GetCSMotif();
    matrixIO->WriteMatrix(pMotif);
    pMotif = (*iterC)->GetWidthMotif();
    if (pMotif != NULL)
      matrixIO->WriteMatrix(pMotif);
  }
  return;
}
