#include "fuzzutilities.h"

namespace FUZZINCLUSIVE {

/******************************************************************************
  Description:  TranslateToSequence() (acgt->0123) 
  Author_Date:  MC_2009/07/27
******************************************************************************/
Sequence *
TranslateToSequence(string * code)
{
  Sequence * site = new Sequence;
  int i = 0;
  while ( i < (int)code->length())
  {
    switch (code->at(i))
    {
    case 'a':
    case 'A':
      site->push_back(0);
      break;
    case 'c':
    case 'C':
      site->push_back(1);
      break;
    case 'g':
    case 'G':
      site->push_back(2);
      break;
    case 't':
    case 'T':
      site->push_back(3);
      break;
    default:
      site->push_back(-1);
    }
    i++;
  }
  return site;
}
/******************************************************************************
  Description:  BestShift : return shift with maximal matches, minimal shift 
                and at least 50% overlap with Inst1 (main instance)
  Author_Date:  MC_2009/07/27
******************************************************************************/
int
BestShift(Instance * Inst1, Instance * Inst2)
{
  //cerr << "debug::fuzzutilities::BestShift- begin" << endl;

  int bestShift(0), bestMatch(INT_MIN), match(0);
  int L1(Inst1->Length()), L2(Inst2->Length()), minOverlap(ceil(L1/2)); 
  int a1, a2, leftshift, rightshift;
  Sequence * inst1 = Inst1->Site();   Sequence * inst2 = Inst2->Site(); 

  // scan the shifting area (shift = against Inst1)
  rightshift = L1 - minOverlap;
  if (rightshift + L2 < L1) rightshift = L1 - L2; // short I2
  leftshift = minOverlap - L2; // negative value
  if (leftshift > 0) leftshift = 0; // short I2

  for(int shift = leftshift; shift <= rightshift; shift++)
  { 
    // set the borders of overlapping region
    if (shift <= 0) a1 = 0; else a1 = shift;
    if (shift + L2 < L1) a2 = shift + L2; else a2 = L1;
    // count matches in (a1-a2)
    if ((a2-a1) <= bestMatch) continue;
    match = 0;
    for(int p = a1; p < a2; p++)
    {
      if ( (*inst1)[p] != -1 && (*inst1)[p] == (*inst2)[p-shift] )
        match++;
    }
    // store best
    if (match >= bestMatch)
    {
      if (match > bestMatch) 
      { bestMatch = match; bestShift = shift;}
      else
      { // same match-score -> take smallest shift
        if ( fabs(shift) < fabs(bestShift))
        { bestMatch = match; bestShift = shift;}
      }
    }
  }
  //cerr << "debug::fuzzutilities::BestShift- end" << endl;
  return bestShift;
}
/******************************************************************************
  Description:  create Sequence with nucleotide information of set length, 
                given the shift against zero
  Author_Date:  MC_2009/07/27
******************************************************************************/
Sequence *
ConvertSiteToLength(Instance * Inst2, int L1, int shift)
{
  //cerr << "debug fuzzutilties::ConvertSiteToLength - begin" << endl;
  // start with a string (convert to Sequence at end)
  string * pString2 = new string(*Inst2->PrintSite());

  int start2(Inst2->Start()), L2(Inst2->Length());
  Instance *minInst(NULL),*maxInst(NULL);
  int leftshift, rightshift;
  string * code = NULL;

  // check the heading positions
  if (shift <= 0) 
  { // remove heading positions

    pString2->erase(0, -shift);
  }
  else 
  { // look for informative heading nucleotides

    minInst = Inst2->MinStart();
    if (minInst == NULL)
    {
      pString2->insert(0, shift, 'N');
    }
    else 
    { 
      // get maximal information at the left
      leftshift = start2 - minInst->Start();
      code = new string(minInst->PrintSite()->substr(0, leftshift));
      if (leftshift > shift) 
      { 
        pString2->insert(0, *code, leftshift-shift, shift);}
      else 
      { 
        pString2->insert(0, *code, 0, leftshift);
        pString2->insert(0, shift-leftshift, 'N');
      }
      delete code; code = NULL;
    }
    minInst = NULL;
  }

  // check the trailing positions
  if (shift + L2 >= L1) 
  { // remove trailing positions
    pString2->erase(L1, shift + L2 - L1);
  }
  else 
  { // look for informative right positions
    maxInst = Inst2->MaxEnd();
    if (maxInst == NULL) 
    { 
      pString2->append(L1 - shift - L2, 'N');
    }
    else
    { 
      // get maximal information at the right
      rightshift = maxInst->Start() + maxInst->Length() - start2 - L2;
      code = new string(*(maxInst->PrintSite()));
      *code = code->substr(maxInst->Length()- rightshift, rightshift);
      if (rightshift >= L1 - shift - L2) 
      { 
        pString2->append(*code, 0, L1 - shift - L2);}
      else 
      { 
        pString2->append(*code, 0, rightshift);
        pString2->append(L1 - shift - L2 - rightshift, 'N');
      }
      delete code; code = NULL;
    }
    maxInst = NULL;
  }

  // convert to Sequence format
  Sequence * site = FUZZINCLUSIVE::TranslateToSequence(pString2);

  // cleanup
 delete pString2; pString2 = NULL;

if ( (int)site->size() != L1)
{cerr << "--Error--debug fuzzutilities::ConvertSiteToLength - errorrrrr ==> debug\n";
 delete site; site = NULL;// this will provoke program error
}

  //cerr << "debug fuzzutilties::ConvertSiteToLength - end" << endl;
  return site;
}

/******************************************************************************
  Description:  get motif from pwms in column Start, width = w
  Author_Date:  MC_2009/03/02
******************************************************************************
  PWM * 
  GetMotif(mwArray pwms, const int &pwmStart, const int &w)
  {
    //cerr << "debug::fuzzutilities::GetMotif() - begin" << endl;
    // set zero pseudo (already applied in mwArray)
    double * pseudo = new double[4];
    for(int n = 0; n < 4; n++) pseudo[n] = 0;
    // create matrix 
    double ** matrix = new double*[w];
    for(int j = 0; j < w; j++)
    {
      matrix[j] = new double[4];
      for(int n = 0; n < 4; n++)
        // copy value to matrix
        matrix[j][n] = pwms(n+1, pwmStart+j);
    }
    // create PWM 
    PWM * motif = new PWM(w, matrix, pseudo);
    // cleanup
    delete[] pseudo; pseudo = NULL;
    for(int i = 0; i < w; i++)
      delete[] matrix[i];
    delete[] matrix; matrix = NULL;
    //cerr << "debug::fuzzutilities::GetMotif() - end" << endl;
    return motif;
  }
  
  */
/******************************************************************************
  Description:  mwArrayToPWM
  Author_Date:  MC_2009/02/19
******************************************************************************
  PWM *
  mwArrayToPWM(mwArray pwm, const int &w)
  {
    // set zero pseudo (already applied in mwArray)
    double * pseudo = new double[4];
    for(int n = 0; n < 4; n++) pseudo[n] = 0;
    // create matrix 
    double ** matrix = new double*[w];
    for(int j = 0; j < w; j++)
    {
      matrix[j] = new double[4];
      for(int n = 0; n < 4; n++)
        // copy value to matrix
        matrix[j][n] = pwm.Get(1, n+1,j+1);
    }
    // create PWM 
    PWM * motif = new PWM(w, matrix, pseudo);
    return motif;
  }
*/

} // end of the namespace
